/*
 * Copyright (C) 2001-2005 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "NmdcHub.h"

#include "ResourceManager.h"
#include "ClientManager.h"
#include "SearchManager.h"
#include "ShareManager.h"
#include "CryptoManager.h"
#include "ConnectionManager.h"

#include "Socket.h"
#include "UserCommand.h"
#include "ZUtils.h"

NmdcHub::NmdcHub(const string& aHubURL) : Client(aHubURL, '|'), supportFlags(0),  
	state(STATE_CONNECT), lastActivity(GET_TICK()), reconnect(true), lastbytesshared(0), 
	validatenicksent(false), bFirstOpList(true), PtokaX(false), ShitInBox(false) {
	TimerManager::getInstance()->addListener(this);
}

NmdcHub::~NmdcHub() throw() {
	TimerManager::getInstance()->removeListener(this);

	Lock l(cs);
	clearUsers();
}

void NmdcHub::connect() {
	setRegistered(false);
	setReconnDelay(120 + Util::rand(0, 60));
	reconnect = true;
	supportFlags = 0;
	lastmyinfo.clear();
	lastbytesshared = 0;
	validatenicksent = false;
	bFirstOpList = true;
	PtokaX = false;
    ShitInBox = false;

	if(socket->isConnected()) {
		disconnect();
	}

	reloadSettings();

	state = STATE_LOCK;

	if(getPort() == 0) {
		setPort(411);
	}
	socket->connect(getAddress(), getPort());
}

void NmdcHub::connect(const User* aUser) {
	checkstate(); 
	dcdebug("NmdcHub::connectToMe %s\n", aUser->getNick().c_str());
	char buf[256];
	int chars = 0;
	if(SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE) {
		chars = sprintf_s(buf, 256, "$ConnectToMe %s %s:%d|", toNmdc(aUser->getNick()).c_str(), getLocalIp().c_str(), ConnectionManager::getInstance()->getPort());
		ConnectionManager::iConnToMeCount++;
	} else {
		chars = sprintf_s(buf, 256, "$RevConnectToMe %s %s|", toNmdc(getNick()).c_str(), toNmdc(aUser->getNick()).c_str());
		Lock l(cs);
		User::NickIter i = users.find(aUser->getNick());
		if(i != users.end()) {
			User::Ptr u = i->second;
			u->setRCTMSend(true);
		}
	}
	if(chars != -1) send(buf, chars); 
}

int64_t NmdcHub::getAvailable() const {
	Lock l(cs);
	int64_t x = 0;
	for(User::NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
		x+=i->second->getBytesShared();
	}
	return x;
}

void NmdcHub::refreshUserList(bool unknownOnly /* = false */) {
	Lock l(cs);
	if(unknownOnly) {
		for(User::NickIter i = users.begin(); i != users.end(); ++i) {
			if(i->second->getConnection().empty()) {
				getInfo(i->second);
			}
		}
	} else {
		clearUsers();
		getNickList();
	}
}

void NmdcHub::clearUsers() {
	for(User::NickIter i = users.begin(); i != users.end(); ++i) {
		ClientManager::getInstance()->putUserOffline(i->second);		
	}
	users.clear();
}

void NmdcHub::on(Line, const string& l) throw() {
    lastActivity = GET_TICK();
	if(l[0] == '$') {
		if(l[1] == 'Z' && l[2] == ' ') {
			COMMAND_DEBUG(l, DebugManager::HUB_IN, getIpPort());
			ZLine((char *)l.c_str(), l.size());
			return;
		} else {
			string line = fromNmdc(l);
			char *sLine = (char *)line.c_str();
			COMMAND_DEBUG(sLine, DebugManager::HUB_IN, getIpPort());
			DcLine(sLine, line.size(), (char *)l.c_str(), l.size());
            return;
		}
    } else {
		string line = fromNmdc(l);
		char *sLine = (char *)line.c_str();
		COMMAND_DEBUG(sLine, DebugManager::HUB_IN, getIpPort());
        ChatLine(sLine);
        return;
    }
}

void NmdcHub::ZLine(char* aLine, int iaLineLen) throw() {
    if(iaLineLen < 5) return;
    
    aLine += 3; char *sTemp = aLine; 
    size_t iLen = iaLineLen-3, 
		ipLen = sTemp-aLine;
	bool corrupt = false;

	// unescape \\ to \ and \P to |
	while((sTemp = (char *)memchr(sTemp, '\\', iLen-(ipLen))) != NULL) {
		ipLen = sTemp-aLine;
        if(ipLen < iLen) {
    		switch(sTemp[1]) {
    			case '\\':
                    memmove(sTemp, sTemp+1, iLen-(ipLen));
                    iLen--; sTemp++; ipLen++;
    				continue;
    			case 'P':
                    memmove(sTemp, sTemp+1, iLen-(ipLen));
    				sTemp[0] = '|';
    				iLen--; sTemp++; ipLen++;
    				continue;
    			default:
    				corrupt = true;
    				break;
    		}
        } else {
            corrupt = true;
        }
		break;
	}

	if(!corrupt) {
		// unzip the ZBlock
		UnZFilter filter;
		string lines = "";
		bool more = true;
		string::size_type j, readFromPos = 0, estOutSize = iLen * 4;
		AutoArray<u_int8_t> temp(estOutSize);

		while(more) {
			j = estOutSize;
			try {
				more = filter((const char *)aLine+readFromPos, iLen, temp, j);
			} catch(...) {
				dcdebug("Error during Zline decompression\n");
				break;
			}
			lines += string((char*)(u_int8_t*)temp, j);
			readFromPos += iLen;
			
			char * sLines = (char *)lines.c_str();

			// split lines up into indiviual commands
            while((sTemp = strchr(sLines, '|')) != NULL) {
                sTemp[0] = '\0';
				string line = fromNmdc(sLines);
				char *sLine = (char *)line.c_str();
				// "fire" the line
                COMMAND_DEBUG("ZLine " + string(sLine), DebugManager::HUB_IN, getIpPort());
            	switch(sLine[0]) {
                    case '$':
                        DcLine(sLine, line.size(), sLines, (sTemp-sLines));
						break;
                    default:
                        ChatLine(sLine);
                        break;
                }
        		sLines = sTemp+1;
        	}
        	
        	// last token, roll over if more data to decompress
            if(sLines[0] != NULL)
        	   lines = string(sLines);

			// a nmdc command over 1mb ?
			if(lines.size() > 1048576) {
				dcdebug("Malicious data found during ZLine decompression\n");
				break;
			}
		}
	} else {
		dcdebug("Corrupt Zline datastream\n");
	}
}

void NmdcHub::ChatLine(char* aLine) throw() {
	char *temp;

	// Check if we're being banned...
	if(state != STATE_CONNECTED) {
		if(strstr(aLine, "banned") != 0) {
			reconnect = false;
		}
	}
		
    string nick = Util::emptyString;

	switch(aLine[0]) {
        case '<':
            if((temp = strchr(aLine+1, '>')) != NULL) {
                temp[0] = NULL;
                nick = aLine+1;
                if(temp[1] == ' ' && temp[2] == '/' && tolower(temp[3]) == 'm' && tolower(temp[4]) == 'e' && temp[5] == ' ' && temp[6] != NULL) {
                    for(int iNickLen = (temp - aLine); iNickLen > 1; iNickLen--) {
                        aLine[iNickLen+4] = aLine[iNickLen-1];
                    }
					aLine[5] = ' ';
					aLine[4] = '*';
					aLine += 4;
                } else {
                    temp[0] = '>';
                }
            }
                break;
        case '*':
            if((temp = strchr(aLine+2, ' ')) != NULL) {
                temp[0] = NULL;
                nick = aLine+2;
                temp[0] = ' ';
            }
            break;
            default:
                break;
    }

    User::Ptr u = NULL;
    if(nick != Util::emptyString) {
        Lock l(cs);
        User::NickIter ni = users.find(nick);
        if(ni != users.end()) {
        	u = ni->second;
        }
    }

    fire(ClientListener::Message(), this, u, Util::validateMessage(aLine, true).c_str());
}

void NmdcHub::DcLine(char* aLine, int iaLineLen, char* bLine, int ibLineLen) throw() {
	char *temp, *temp1;

	switch(aLine[1]) {
        case 'S':
        	// $Search
        	if(strncmp(aLine+2, "earch ", 6) == 0) {
        	    if(iaLineLen < 19) return;
        	    
        	    aLine += 8;
        	    
        		if(state != STATE_CONNECTED || (temp = strchr(aLine, ' ')) == NULL || temp[1] == NULL) return;
        
        		temp[0] = NULL; temp += 1;
        		if(aLine[0] == NULL || temp[1] == NULL || temp[2] == NULL || temp[3] == NULL || temp[4] == NULL) return;
        
        		string seeker = aLine;
        		int iseekerLen = (temp-aLine)-1;
        		
        		bool bPassive = (_strnicmp(seeker.c_str(), "Hub:", 4) == 0);
        
        		// We don't wan't to answer passive searches if we're in passive mode...
        		if(bPassive == true && SETTING(CONNECTION_TYPE) != SettingsManager::CONNECTION_ACTIVE) {
        			return;
        		}
        		// Filter own searches
        		if((SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE) && bPassive == false) {
        			if((strcmp(seeker.c_str(), (getLocalIp() + ":" + Util::toString(SearchManager::getInstance()->getPort())).c_str())) == 0) {
        				return;
        			}
        		} else {
        			if(strcmp(seeker.c_str() + 4, getNick().c_str()) == 0) {
        				return;
        			}
        		}
        
        		{
        			Lock l(cs);
        			u_int32_t tick = GET_TICK();
        			seekers.push_back(make_pair(seeker, tick));
        			// First, check if it's a flooder
        			for(FloodIter fi = flooders.begin(); fi != flooders.end(); ++fi) {
        				if(fi->first == seeker)
        					return;
        			}
        			int count = 0;
        			for(FloodIter fi = seekers.begin(); fi != seekers.end(); ++fi) {
        				if(fi->first == seeker)
        					count++;
        
        				if(count > 7) {
        				    // BFU don't need to know search spammers ;)
        				    if(getOp()) {
        						if(bPassive == true && iseekerLen > 4) {
        							fire(ClientListener::SearchFlood(), this, seeker.c_str()+4);
        						} else {
        							fire(ClientListener::SearchFlood(), this, seeker + STRING(NICK_UNKNOWN));
        						}
        					}				
        					flooders.push_back(make_pair(seeker, tick));
        					return;
        				}
        			}
        		}
        		int a;
        		if(temp[0] == 'F') {
        			a = SearchManager::SIZE_DONTCARE;
        		} else {
        			if(temp[2] == 'F') {
        				a = SearchManager::SIZE_ATLEAST;
        			} else {
        				a = SearchManager::SIZE_ATMOST;
        			}
        		}
        		temp += 4;
        		if((temp1 = strchr(temp, '?')) == NULL || temp1[1] == NULL || temp1[2] == NULL) return;
        
        		temp1[0] = NULL; temp1[2] = NULL;
        		if(temp[0] == NULL) return;
        
        		int64_t size = _atoi64(temp);
        		int type = atoi(temp1+1) - 1; temp1 += 3;
        		if(temp1[0] != NULL) {
        			if(bPassive == true && iseekerLen > 4) {
        				User::Ptr u;
        				
        				{
        				    Lock l(cs);
        				    User::NickIter ni = users.find(seeker.c_str()+4);
        				    if(ni != users.end()) {
        					   u = ni->second;
        				    }
                        }
                        
                        if(u && !u->isSet(User::PASSIVE) && u->getMode() != "A") {
        					u->setFlag(User::PASSIVE);
        					updated(u);
        				}
        			}
        			fire(ClientListener::NmdcSearch(), this, seeker, a, size, type, temp1, bPassive);
        		}
        		return;
            }
            
            // $SR
            if(strncmp(aLine+2, "R ", 2) == 0) {
                if(iaLineLen < 5) return;
                
    			SearchManager::getInstance()->onSearchResult(aLine);
    			return;
            }
            
            // $Supports
           	if(strncmp(aLine+2, "upports", 7) == 0) {
        		StringList sl;
        		aLine += 10;
        		if(iaLineLen < 11) {
        			validateNick(getNick());
        			return;
        		}
        
        		while(aLine != NULL) {
                    if((temp = strchr(aLine, ' ')) != NULL) {
                        temp[0] = NULL; temp++;
                    }

        			sl.push_back(aLine);
        			switch(aLine[0]) {
                        case 'U':
                			if((strcmp(aLine+1, "serCommand")) == 0) {
                				supportFlags |= SUPPORTS_USERCOMMAND;
                			} else if((strcmp(aLine+1, "serIP2")) == 0) {
                				supportFlags |= SUPPORTS_USERIP2;
                            }
                            break;
                        case 'N':
                			if((strcmp(aLine+1, "oGetINFO")) == 0) {
                				supportFlags |= SUPPORTS_NOGETINFO;
                            }
                            break;
                        case 'Q':
                			if((strcmp(aLine+1, "uickList")) == 0) {
                				if(state == STATE_HELLO) {
                					state = STATE_MYINFO;
                					updateCounts(false);
                					myInfo();
                					getNickList();
                					supportFlags |= SUPPORTS_QUICKLIST;
                				}
                			}
                			break;
						case 'Z':
							if((strcmp(aLine+1, "Line")) == 0) {
								supportFlags |= SUPPORTS_ZLINE;
							}
							break;
                		default:
                            dcdebug("NmdcHub::onLine Unknown supports %s\n", aLine);
                            break;
                    }
        			aLine = temp;
        		}
        		if(!(getSupportFlags() & SUPPORTS_QUICKLIST)) {
        			validateNick(getNick());
        		}
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
    	case 'M':
        	// $MyINFO $ALL  $ $$$$
        	if(strncmp(aLine+2, "yINFO ", 6) == 0) {
                if(iaLineLen < 21) return;
                
        	    char *Description, *Tag, *Connection, *Email, *Share;
        	    aLine += 13;
        	    if((Description = strchr(aLine, ' ')) == NULL || Description[1] == NULL)
        	         return;
        
        		Description[0] = NULL;
        		Description += 1;
        
        	    if(aLine[0] == NULL)
        	         return;
        
           		User::Ptr u;
        		string nick = aLine;
        
        		{
        			Lock l(cs);
        			User::NickIter ni = users.find(nick);
        			if(ni == users.end()) {
        				u = users[nick] = ClientManager::getInstance()->getUser(nick, this);
        			} else {
        				u  = ni->second;
        			}
        		}
        
        		// New MyINFO received, delete all user variables... if is bad MyINFO is user problem not my ;-)
        		u->setNick(nick);
        		u->setBytesShared(0);
        		u->setDescription(Util::emptyString);
        		u->setcType(9);
        		u->setConnection(Util::emptyString);
        		u->setEmail(Util::emptyString);
        		u->setTag(Util::emptyString);
        		u->setClientID(Util::emptyString);
        		u->setVersion(Util::emptyString);
        		u->setMode(Util::emptyString);
        		u->setHubs(Util::emptyString);
        		u->setSlots(Util::emptyString);
        		u->setUpload(Util::emptyString);
                             		
        	    Connection = strchr(Description, '$');
        	    if(Connection && Connection+1 && Connection+2 && Connection+3) {
        			Connection[0] = NULL;
        			Connection += 1;
        			if(Description) {
        				if(*(Connection-2) == '>') {
        					if((Tag = strrchr(Description, '<')) != NULL) {
        						u->setTag(Tag);
        						u->TagParts();
        						Tag[0] = NULL;
        					} else
        						u->setTag(Util::emptyString);
        				}
        				u->setDescription(Util::validateMessage(Description, true));
        			}
        			Email = strchr(Connection+2, '$');
        			if(Email && Email+1) {
        				Email[0] = NULL;
        				Email += 1;
        				Share = strchr(Email, '$');
        				if(Share && Share+1) {
        					if((temp = strchr(Share+1, '$')) != NULL) temp[0] = NULL;
        					
        					Share[0] = NULL;
        					Share += 1;
        					u->setEmail(Util::validateMessage(Email, true));
        					u->setBytesShared(Share);
        				}
        				Connection[1] = NULL;
        				if(Connection[0] != ' ')
        					u->setMode(Connection);			    
        
        				Connection += 2;
        				char status = *(Email-2);
        				*(Email-2) = NULL;
       					switch(status) {
                            case 1:
                                u->unsetFlag(User::AWAY);
      					        u->unsetFlag(User::SERVER);
       					        u->unsetFlag(User::FIREBALL);
                                break;
                            case 2:
                            case 3:
          						u->setFlag(User::AWAY);
           						u->unsetFlag(User::SERVER);
           						u->unsetFlag(User::FIREBALL);
           						break;
       					    case 4:
                            case 5:
           						u->setFlag(User::SERVER);
           						u->unsetFlag(User::AWAY);
           						u->unsetFlag(User::FIREBALL);
           						break;
       					    case 6:
                            case 7:
           						u->setFlag(User::SERVER);
           						u->setFlag(User::AWAY);
           						u->unsetFlag(User::FIREBALL);
           						break;
           					case 8:
                            case 9:
           						u->setFlag(User::FIREBALL);
           						u->unsetFlag(User::AWAY);
           						u->unsetFlag(User::SERVER);
           						break;
           					case 10:
                            case 11:
           						u->setFlag(User::FIREBALL);
           						u->setFlag(User::AWAY);
           						u->unsetFlag(User::SERVER);
           						break;
           					default:
                                // WTF ?!? ok go to default...
                                u->unsetFlag(User::AWAY);
      					        u->unsetFlag(User::SERVER);
       					        u->unsetFlag(User::FIREBALL);
                                break;
       					}
        				u->setConnection(Connection);
        				switch(Connection[0]) {
                            case '2':
                                if(strcmp(Connection+1, "8.8Kbps") == 0) {
                                    u->setcType(1);
                                }
                                break;
                            case '3':
            				    if(strcmp(Connection+1, "3.6Kbps") == 0) {
                                    u->setcType(1);
                                }
                                break;
                            case '5':
            					if(strcmp(Connection+1, "6Kbps") == 0) {
                                    u->setcType(1);
                                }
                                break;
                            case 'M':
                                if(strcmp(Connection+1, "odem") == 0) {
            					   u->setcType(1);
                                } else if(strcmp(Connection+1, "icrowave") == 0) {
            					   u->setcType(3);
                                }
                                break;
                            case 'I':
                				if(strcmp(Connection+1, "SDN") == 0) {
            					   u->setcType(2);
                                }
                                break;
                            case 'S':
                				if(strcmp(Connection+1, "atellite") == 0) {
                                    u->setcType(3);
                                }
                                break;
                            case 'W':
                                if(strcmp(Connection+1, "ireless") == 0) {
            					   u->setcType(3);
                                }
                                break;
                            case 'D':
                				if(strcmp(Connection+1, "SL") == 0) {
            					   u->setcType(4);
                                }
                                break;
                            case 'C':
                				if(strcmp(Connection, "Cable") == 0) {
            					   u->setcType(5);
                                }
                                break;
                            case 'L':
                				if(strcmp(Connection+1, "AN(T1)") == 0 || strcmp(Connection+1, "AN(T3)") == 0) {
            					   u->setcType(6);
                                }
                                break;
            				default:
            					u->setcType(9);
            					break;
                        }
        			}
        		}
        
        		if(state == STATE_MYINFO) {
        			if(_stricmp(u->getNick().c_str(), getNick().c_str()) == NULL) {
        				setMe(u);
        				state = STATE_CONNECTED;
        				updateCounts(false);
        				u->setFlag(User::DCPLUSPLUS);
        				if(SETTING(CONNECTION_TYPE) != SettingsManager::CONNECTION_ACTIVE)
        					u->setFlag(User::PASSIVE);
        				else
        					u->unsetFlag(User::PASSIVE);
        			}
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'Q':
        	// $Quit
    		if(strncmp(aLine+2, "uit ", 4) == 0) {
                if(iaLineLen < 7) return;
                
        		aLine += 6;
        
        		User::Ptr u;
        		{
        			Lock l(cs);
        			User::NickIter i = users.find(aLine);
        			if(i == users.end()) {
        				dcdebug("C::onLine Quitting user %s not found\n", aLine);
        				return;
        			}
        			
        			u = i->second;
        			users.erase(i);
        		}
        		
        		fire(ClientListener::UserRemoved(), this, u);
        		ClientManager::getInstance()->putUserOffline(u, true);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'C':
    		// $ConnectToMe
    		if(strncmp(aLine+2, "onnectToMe ", 11) == 0) {
                if(state != STATE_CONNECTED || iaLineLen < 16) return;
                
        		aLine += 13;
        		if(state != STATE_CONNECTED || (temp = strchr(aLine, ' ')) == NULL || temp[1] == NULL) return;
        
        		temp[0] = NULL; temp += 1;
        		if(aLine[0] == NULL) return;
        
        		if(_stricmp(aLine, getNick().c_str()) != 0) // Check nick... is CTM really for me ? ;o)
        			return;
        
        		if((temp1 = strchr(temp, ':')) == NULL || temp1[1] == NULL) return;
        
        		temp1[0] = NULL; temp1 += 1;
        		if(temp[0] == NULL) return;
        
        		ConnectionManager::getInstance()->nmdcConnect(temp, (short)atoi(temp1), getNick());
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'R':
        	// $RevConnectToMe
        	if(strncmp(aLine+2, "evConnectToMe ", 14) == 0) {
        		if(state != STATE_CONNECTED || iaLineLen < 19) return;
        
        		aLine += 16;
        		if((temp = strchr(aLine, ' ')) == NULL || temp[1] == NULL) return;

        		temp[0] = NULL; temp += 1;
        		if(aLine[0] == NULL) return;
        		
        		if(_stricmp(temp, getNick().c_str()) != 0) // Check nick... is RCTM really for me ? ;-)
        			return;
        		
        		User::Ptr u;
        		
        		{
        			Lock l(cs);

        			User::NickIter i = users.find(aLine);
        
        			if(i == users.end()) {
        				return;
        			}
        
        			u = i->second;
        		}
        
        		if(u) {
        			if(SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE) {
        				connectToMe(u);
        			} else {
        				// Notify the user that we're passive too...
        				if(!u->isSet(User::PASSIVE) && u->getRCTMSend() == false) {
        					revConnectToMe(u);
                        }
                        u->setRCTMSend(false);
        			}
        
                	if(!u->isSet(User::PASSIVE)) {
        				u->setFlag(User::PASSIVE);
        				updated(u);
        			}
        		}
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'H':
        	// $HubName
        	if(strncmp(aLine+2, "ubName ", 7) == 0) {
                if(iaLineLen < 10) return;
                
        		aLine += 9;
        
        		name = aLine;
        		if((temp = strstr(aLine, " - ")) != NULL) {
        		    temp[0] = NULL;
        			notopicname = aLine;
        		} else {
        			notopicname = name;
        		}    			
        		fire(ClientListener::HubUpdated(), this);
        		return;
            }
            
            // $Hello
           	if(strncmp(aLine+2, "ello ", 5) == 0) {
        		if(getSupportFlags() & SUPPORTS_QUICKLIST || iaLineLen < 8) return;
        		aLine += 7;
        		if(aLine[0] == NULL) return;
        
        		string nick = aLine;
        		User::Ptr u = ClientManager::getInstance()->getUser(nick, this);
        		{
        			Lock l(cs);
        			users[nick] = u;
        		}
        		
        		u->setNick(nick);
        		
        		if(state == STATE_HELLO) {
            		if(_stricmp(u->getNick().c_str(), getNick().c_str()) == NULL) {
            			setMe(u);
                        state = STATE_CONNECTED;
        			    updateCounts(false);
            			u->setFlag(User::DCPLUSPLUS);
            			if(SETTING(CONNECTION_TYPE) != SettingsManager::CONNECTION_ACTIVE)
            				u->setFlag(User::PASSIVE);
            			else
            				u->unsetFlag(User::PASSIVE);
            		}
        
        			version();
        			if(ShitInBox == true || getStealth() == true) {
        				getNickList();
        				myInfo();
                    } else {
        				myInfo();
        				getNickList();
                    }
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }
            
            // $HubIsFull
           	if(strcmp(aLine+2, "ubIsFull") == 0) {
				fire(ClientListener::HubFull(), this);
    			return;
        	}
        	
        	dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'U':
        	// $UserCommand
        	if(strncmp(aLine+2, "serCommand ", 11) == 0) {
                if(iaLineLen < 16) return;
                
        		aLine += 13;
        		if((temp = strchr(aLine, ' ')) == NULL || temp[1] == NULL) return;
        
        		temp[0] = NULL;
        		int type = atoi(aLine);
        		int ctx = 0;
        		
        		if(type == 0 || type == 255) {
        			ctx = atoi(temp+1);
        		} else {
        			temp1 = temp+1;
        			if((temp = strchr(temp1, ' ')) == NULL || temp[1] == NULL) return;
	        		temp[0] = NULL;
	        		ctx = atoi(temp1);
	        		temp = temp+1;
        		}
        
        		switch(type) {
                    case UserCommand::TYPE_SEPARATOR:
                    case UserCommand::TYPE_CLEAR:
        			    fire(ClientListener::UserCommand(), this, type, ctx, Util::emptyString, Util::emptyString);
        			    break;
            		case UserCommand::TYPE_RAW:
                    case UserCommand::TYPE_RAW_ONCE:
        			    if((temp1 = strchr(temp, '$')) == NULL || temp1[1] == NULL) return;
        
        			    temp1[0] = NULL; temp1 += 1;
        			    if(temp[0] == NULL) return;
        
        			    fire(ClientListener::UserCommand(), this, type, ctx, Util::validateMessage(temp, true, false), Util::validateMessage(temp1, true, false));
        			    break;
        			default:
                        // ?!?
                        break;
        		}
        		return;
            }
            
            // $UserIP
           	if(strncmp(aLine+2, "serIP ", 6) == 0) {
                if(iaLineLen < 17) return;
                
        		User::List v;
        		StringList l;
        		aLine += 8;
        
        		while((temp = strchr(aLine, ' ')) != NULL && temp[1] != NULL) {
        			temp[0] = NULL; temp += 1;
        			if(aLine[0] == NULL) break;
        			
        			if((temp1 = strchr(temp, '$')) != NULL) {
            			temp1[0] = NULL;
            			if(temp[0] == NULL) break;
                    }
            			
                	User::Ptr u = NULL;
                
                	{
                		Lock l(cs);
                		User::NickIter ni = users.find(aLine);
                		if(ni != users.end()) {
                			u  = ni->second;
                		}
                	}

                    if(u) {
                        v.push_back(u);
                        v.back()->setIp(temp);
                    }
                    
                    if(temp1 == NULL) {
                        break;
                    }

                    aLine = temp1+2;
        		}
				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'L':
        	// $Lock
        	if(strncmp(aLine+2, "ock ", 4) == 0) {
        		if(state != STATE_LOCK || ibLineLen < 7) return;
        		
        		bLine += 6;

        		state = STATE_HELLO;
        		if((temp = strstr(bLine, " Pk=")) != NULL && temp[1] != NULL) {
        			temp[0] = NULL; temp += 1;
                }

        		if(bLine[0] == NULL) return;

        		if(temp[0] != NULL) {
        			if(_stricmp(temp+3, "YnHub") == 0) {
        				ShitInBox = true;
        			} else if(strcmp(temp+3, "PtokaX") == 0) {
        				PtokaX = true;
        			}
                }

        		if(CryptoManager::getInstance()->isExtended(bLine)) {
        			StringList feat;
        			feat.push_back("UserCommand"); feat.push_back("NoGetINFO");
            		feat.push_back("NoHello"); feat.push_back("UserIP2");
        			if(getStealth() == false) {
            			feat.push_back("QuickList"); feat.push_back("ZLine");
                    } else {
        				feat.push_back("TTHSearch");

        				if(BOOLSETTING(COMPRESS_TRANSFERS))
        					feat.push_back("GetZBlock");
                    }

        			supports(feat);
					if(PtokaX == false || getStealth() == true) key(CryptoManager::getInstance()->makeKey(bLine));
        		} else {
        			key(CryptoManager::getInstance()->makeKey(bLine));
        			validateNick(getNick());
        		}
                return;
    		}

    		// $LogedIn
    		if(strncmp(aLine+2, "ogedIn", 6) == 0) {
    			// nothing to do here...
    			return;
            }

            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
    	case 'F':
            // $ForceMove
        	if(strncmp(aLine+2, "orceMove ", 9) == 0) {
                if(iaLineLen < 12) return;
                
        		aLine += 11;
        
        		disconnect();
        		fire(ClientListener::Redirect(), this, aLine);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'V':
        	// $ValidateDenide
        	if(strncmp(aLine+2, "alidateDenide", 13) == 0) {
        		disconnect();
				fire(ClientListener::NickTaken(), this);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'N':
        	// $NickList
        	if(strncmp(aLine+2, "ickList ", 8) == 0) {
                if(iaLineLen < 11) return;
                
        		User::List v;
        		aLine += 10;
        
        		while((temp = strchr(aLine, '$')) != NULL) {
        			temp[0] = NULL;
        			if(aLine[0] == NULL) break;
        
        			v.push_back(ClientManager::getInstance()->getUser(aLine, this));
        			aLine = temp+2;
        		}
        
        		{
        			Lock l(cs);
        			for(User::Iter it2 = v.begin(); it2 != v.end(); ++it2) {
        				users[(*it2)->getNick()] = *it2;
        			}
        		}
        		
        		if(!(getSupportFlags() & SUPPORTS_NOGETINFO)) {
        			string tmp;
        			// Let's assume 10 characters per nick...
        			tmp.reserve(v.size() * (11 + 10 + getNick().length())); 
        				string n = ' ' +  toNmdc(getNick()) + '|';
        			for(User::List::const_iterator i = v.begin(); i != v.end(); ++i) {
        				tmp += "$GetINFO ";
        					tmp += toNmdc((*i)->getNick());
        					tmp += n;
        			}
        			if(!tmp.empty()) {
        				send(tmp);
        			}
        		}
				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'O':
        	// $OpList
        	if(strncmp(aLine+2, "pList ", 6) == 0) {
                if(iaLineLen < 9) return;
                
        		User::List v;
        		aLine += 8;
        
        		while((temp = strchr(aLine, '$')) != NULL) {
        			temp[0] = NULL;
        			if(aLine[0] == NULL) break;
        
        			v.push_back(ClientManager::getInstance()->getUser(aLine, this));
        			v.back()->setFlag(User::OP);
        			aLine = temp+2;
        		}
        
        		{
        			Lock l(cs);
        			for(User::Iter it2 = v.begin(); it2 != v.end(); ++it2) {
        				users[(*it2)->getNick()] = *it2;
        			}
        		}
        		fire(ClientListener::UsersUpdated(), this, v);
        		updateCounts(false);
        		if(bFirstOpList == true) {
        			if(getRegistered() == true) {
        				myInfo();
                    }
        			bFirstOpList = false;
        		}
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'T':
        	// $To:
        	if(strncmp(aLine+2, "o: ", 3) == 0) {
                if(iaLineLen < 17) return;
                
        		if((temp1 = strstr(aLine+5, "From:")) != NULL && strlen(temp1) > 6) {
        			if((temp = strchr(temp1+6, '$')) == NULL || temp[1] == NULL) return;
        
        			temp1 += 6; *(temp-1) = NULL; temp += 1;
        			if(temp1[0] == NULL || temp[0] == NULL) return;

               		User::Ptr u;
            
            		{
            			Lock l(cs);
            			User::NickIter ni = users.find(temp1);
            			if(ni == users.end()) {
            				u = ClientManager::getInstance()->getUser(temp1, this, false);
            			} else {
            				u  = ni->second;
            			}
            		}
                        		
        			fire(ClientListener::PrivateMessage(), this, u, Util::validateMessage(temp, true));
        		}
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'G':
        	// $GetPass
        	if(strcmp(aLine+2, "etPass") == 0) {
        		setRegistered(true);
        		fire(ClientListener::GetPassword(), this);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
        case 'B':
        	// $BadPass
        	if(strcmp(aLine+2, "adPass") == 0) {
        		fire(ClientListener::BadPassword(), this);
        		return;
            }
            
            dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
            return;
    	default:
		  dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
		  return;
	}
}

string NmdcHub::checkNick(const string& aNick) {
	string tmp = aNick;
	string::size_type i = 0;
	while( (i = tmp.find_first_of("|$ ", i)) != string::npos) {
		tmp[i++]='_';
	}
	return tmp;
}

string NmdcHub::getHubURL() {
	return getAddressPort();
}

void NmdcHub::myInfo() {
	if(state != STATE_CONNECTED && state != STATE_MYINFO) {
		return;
	}
	
	dcdebug("MyInfo %s...\n", getNick().c_str());
	char StatusMode = '\x01';
	string tmp1 = ";**\x1fU9";
	string tmp2 = "+L9";
	string tmp3 = "+G9";
	string tmp4 = "+R9";
	string tmp5 = "+K9";
	string::size_type i;
	
	for(i = 0; i < 6; i++) {
		tmp1[i]++;
	}
	for(i = 0; i < 3; i++) {
		tmp2[i]++; tmp3[i]++; tmp4[i]++; tmp5[i]++;
	}
	char modeChar = '?';
	switch(SETTING(CONNECTION_TYPE)) {
        case SettingsManager::CONNECTION_ACTIVE:
    		modeChar = 'A';
    		break;
    	case SettingsManager::CONNECTION_PASSIVE:
		    modeChar = 'P';
		    break;
    	case SettingsManager::CONNECTION_SOCKS5:
		    modeChar = '5';
		    break;
		default:
            // ?!?
            break;
    }
            
	
	char tag[256];
	string connection = SETTING(CONNECTION);
	if (SETTING(THROTTLE_ENABLE) && SETTING(MAX_UPLOAD_SPEED_LIMIT) != 0) {
		sprintf(tag, "%s%s%s%c%s%s%s%s%s%s>", tmp1.c_str(), DCPPVERSIONSTRING, tmp2.c_str(), modeChar, tmp3.c_str(),
		getCounts().c_str(), tmp4.c_str(), Util::toString(Util::getSlots()).c_str(), tmp5.c_str(), Util::toString(SETTING(MAX_UPLOAD_SPEED_LIMIT)).c_str());
	} else {
		sprintf_s(tag, 256, "%s%s%s%c%s%s%s%s>", tmp1.c_str(), DCPPVERSIONSTRING, tmp2.c_str(), modeChar, tmp3.c_str(),
		getCounts().c_str(), tmp4.c_str(), Util::toString(Util::getSlots()).c_str());
	}
	if(getStealth() == false) {
		if (UploadManager::getFireballStatus()) {
			StatusMode += 8;
		} else if (UploadManager::getFileServerStatus()) {
			StatusMode += 4;
		}

		if(Util::getAway()) {
			StatusMode += 2;
		}
	} else {
		if (connection == "Modem") { connection = "56Kbps"; }
		if (connection == "Wireless") { connection = "Satellite"; }
	}
	char myinfo[512];
	sprintf_s(myinfo, 512, "$MyINFO $ALL %s %s%s$ $%s%c$%s$", toNmdc(getNick()).c_str(), 
		toNmdc(Util::validateMessage(getDescription(), false)).c_str(), tag, connection.c_str(), StatusMode, 
		toNmdc(Util::validateMessage(SETTING(EMAIL), false)).c_str());
	int64_t newbytesshared = ShareManager::getInstance()->getShareSize();
	if(strcmp(myinfo, lastmyinfo.c_str()) != 0 || newbytesshared < (lastbytesshared - 1048576) || newbytesshared > (lastbytesshared + 1048576)) {
		lastmyinfo = myinfo;
		lastbytesshared = newbytesshared;
		tag[0] = NULL;
		sprintf_s(tag, 256, "%s$|", Util::toString(newbytesshared).c_str());
		strcat_s(myinfo, 512, tag);
		send(myinfo);
	}
}

void NmdcHub::disconnect() throw() {	
	state = STATE_CONNECT;
	Client::disconnect();
	{ 
		Lock l(cs);
		clearUsers();
	}
}

void NmdcHub::search(int aSizeType, int64_t aSize, int aFileType, const string& aString){
	checkstate(); 
	AutoArray<char> buf((char*)NULL);
	char c1 = (aSizeType == SearchManager::SIZE_DONTCARE) ? 'F' : 'T';
	char c2 = (aSizeType == SearchManager::SIZE_ATLEAST) ? 'F' : 'T';
	string tmp = Util::validateMessage(toNmdc((aFileType == SearchManager::TYPE_TTH) ? "TTH:" + aString : aString), false);
	string::size_type i;
	while((i = tmp.find(' ')) != string::npos) {
		tmp[i] = '$';
	}
	int chars = 0;
	if((SETTING(CONNECTION_TYPE) == SettingsManager::CONNECTION_ACTIVE) && (!BOOLSETTING(SEARCH_PASSIVE))) {
		string x = getLocalIp();
		buf = new char[x.length() + aString.length() + 64];
		chars = sprintf_s(buf, x.length() + aString.length() + 64, "$Search %s:%d %c?%c?%I64d?%d?%s|", x.c_str(), (int)SearchManager::getInstance()->getPort(), c1, c2, aSize, aFileType+1, tmp.c_str());
	} else {
		buf = new char[getNick().length() + aString.length() + 64];
		chars = sprintf_s(buf, getNick().length() + aString.length() + 64, "$Search Hub:%s %c?%c?%I64d?%d?%s|", toNmdc(getNick()).c_str(), c1, c2, aSize, aFileType+1, tmp.c_str());
	}
	send(buf, chars);
}

// TimerManagerListener
void NmdcHub::on(TimerManagerListener::Second, u_int32_t aTick) throw() {
	if(socket && (lastActivity + getReconnDelay() * 1000) < aTick) {
		// Nothing's happened for ~120 seconds, check if we're connected, if not, try to connect...
		lastActivity = aTick;
		// Try to send something for the fun of it...
		if(isConnected()) {
			dcdebug("Testing writing...\n");
			socket->write("|", 1);
		} else {
			// Try to reconnect...
			if(reconnect && !getAddress().empty())
				connect();
		}
	}
	{
		Lock l(cs);
		
		while(!seekers.empty() && seekers.front().second + (5 * 1000) < aTick) {
			seekers.pop_front();
		}
		
		while(!flooders.empty() && flooders.front().second + (120 * 1000) < aTick) {
			flooders.pop_front();
		}
	}
}

// BufferedSocketListener
void NmdcHub::on(BufferedSocketListener::Failed, const string& aLine) throw() {
	{
		Lock l(cs);
		clearUsers();
	}
	
	if(state == STATE_CONNECTED)
		state = STATE_CONNECT;

	fire(ClientListener::Failed(), this, aLine); 
	updateCounts(true);
}

/* Some parts Copyright (C) 2002-2006 Petr Kozelka, PPK at czdc dot org */
