/* 
 * Copyright (C) 2001-2005 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef NMDC_HUB_H
#define NMDC_HUB_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "TimerManager.h"
#include "SettingsManager.h"

#include "ClientManager.h"

#include "BufferedSocket.h"
#include "User.h"
#include "CriticalSection.h"
#include "Text.h"
#include "ConnectionManager.h"
#include "UploadManager.h"

class NmdcHub : public Client, private TimerManagerListener, private Flags
{
	friend class ClientManager;
public:
	typedef NmdcHub* Ptr;
	typedef list<Ptr> List;
	typedef List::iterator Iter;

	enum SupportFlags {
		SUPPORTS_USERCOMMAND = 0x01,
		SUPPORTS_NOGETINFO = 0x02,
		SUPPORTS_USERIP2 = 0x04,
		SUPPORTS_QUICKLIST = 0x08,
		SUPPORTS_ZLINE = 0x10
	};

#define checkstate() if(state != STATE_CONNECTED) return

	virtual void connect(const User* aUser);
	virtual void hubMessage(const string& aMessage) {
		checkstate();
		char buf[256];
		sprintf_s(buf, 256, "<%s> ", getNick().c_str());
		send(toNmdc(string(buf)+Util::validateChatMessage(aMessage)+"|"));
	}
	virtual void privateMessage(const User* aUser, const string& aMessage) { privateMessage(aUser->getNick(), string("<") + getNick() + "> " + aMessage); }
	virtual void send(const string& a) throw() {
		lastActivity = GET_TICK();
		//dcdebug("Sending %d to %s: %.40s\n", a.size(), getName().c_str(), a.c_str());
		COMMAND_DEBUG(a, DebugManager::HUB_OUT, getIpPort());
		socket->write(a);
	}
	virtual void sendUserCmd(const string& aUserCmd) throw() {
		send(toNmdc(aUserCmd));
	}	
	virtual void search(int aSizeType, int64_t aSize, int aFileType, const string& aString);
	virtual void password(const string& aPass) { send("$MyPass " + toNmdc(aPass) + "|"); }
	virtual void info() { myInfo(); }
	
	virtual size_t getUserCount() const {  Lock l(cs); return users.size(); }
	virtual int64_t getAvailable() const;
	virtual const string& getName() const { return BOOLSETTING(STRIP_TOPIC) ? notopicname : name; };
	virtual const string& getFullName() const { return name; };
	virtual bool getOp() const { return getMe() ? getMe()->isSet(User::OP) : false; };
	
	virtual User::NickMap& lockUserList() throw() { cs.enter(); return users; };
	virtual void unlockUserList() throw() { cs.leave(); };

	virtual string escape(string const& str) const { return Util::validateMessage(str, false); };

	void disconnect() throw();
	void myInfo();
	
	void refreshUserList(bool unknownOnly = false);

	void validateNick(const string& aNick) {
		if (validatenicksent == false) {
			send("$ValidateNick " + toNmdc(aNick) + "|");
			validatenicksent = true;
		}
	};
	void key(const string& aKey) { send("$Key " + aKey + "|"); };	
	void version() { send("$Version 1,0091|"); };
	void getNickList() {
		if(state == STATE_CONNECTED || state == STATE_MYINFO) {
			send("$GetNickList|");
		}
	};
	void getInfo(User::Ptr aUser) {
		 checkstate();
		 char buf[256];
		 int chars = sprintf_s(buf, 256, "$GetINFO %s %s|", toNmdc(aUser->getNick()).c_str(), toNmdc(getNick()).c_str());
		 if(chars != -1) send(buf, chars);
	};
	void getInfo(User* aUser) {
		checkstate();
		char buf[256];
		int chars = sprintf_s(buf, 256, "$GetINFO %s %s|", toNmdc(aUser->getNick()).c_str(), toNmdc(getNick()).c_str());
		if(chars != -1) send(buf, chars);
	};

	void connectToMe(const User::Ptr& aUser) {
		checkstate(); 
		dcdebug("NmdcHub::connectToMe %s\n", aUser->getNick().c_str());
		char buf[256];
		int chars = sprintf_s(buf, 256, "$ConnectToMe %s %s:%d|", toNmdc(aUser->getNick()).c_str(), getLocalIp().c_str(), SETTING(IN_PORT));
		if(chars != -1) send(buf, chars);
		ConnectionManager::iConnToMeCount++;
	}

	void privateMessage(const User::Ptr& aUser, const string& aMessage) {
		privateMessage(aUser->getNick(), string("<") + getNick() + "> " + aMessage);
	}
	void privateMessage(const string& aNick, const string& aMessage) {
		checkstate();
		char buf[512];
		sprintf_s(buf, 512, "$To: %s From: %s $", toNmdc(aNick).c_str(), toNmdc(getNick()).c_str());
		send(string(buf)+toNmdc(Util::validateChatMessage(aMessage))+"|");
	}
	void supports(const StringList& feat) { 
		string x;
		for(StringList::const_iterator i = feat.begin(); i != feat.end(); ++i) {
			x+= ' ' + *i;
		}
		if(getStealth() == false) {
		  send("$Supports" + x + '|');
        } else {
		  send("$Supports" + x + " |");
        }
	}
	void revConnectToMe(const User::Ptr& aUser) {
		checkstate(); 
		dcdebug("NmdcHub::revConnectToMe %s\n", aUser->getNick().c_str());
		char buf[256];
		int chars = sprintf_s(buf, 256, "$RevConnectToMe %s %s|", toNmdc(getNick()).c_str(), toNmdc(aUser->getNick()).c_str());
		User::NickIter i = users.find(aUser->getNick());
		if(i != users.end()) {
			User::Ptr u = i->second;
			u->setRCTMSend(true);
		}
		if(chars != -1) send(buf, chars);
	}

	void send(const char* aBuf, int aLen) throw() {
		lastActivity = GET_TICK();
		COMMAND_DEBUG((string)aBuf, DebugManager::HUB_OUT, getIpPort());
		socket->write(aBuf, aLen);
	}

	GETSET(int, supportFlags, SupportFlags);
private:
	enum States {
		STATE_CONNECT,
		STATE_LOCK,
		STATE_HELLO,
		STATE_MYINFO,
		STATE_CONNECTED
	} state;

	string name;
	string notopicname;

	u_int32_t lastActivity;

	mutable CriticalSection cs;

	User::NickMap users;

	bool reconnect;
	string lastmyinfo;
	bool validatenicksent, bFirstOpList;
	int64_t lastbytesshared;
	bool PtokaX, ShitInBox;

	typedef list<pair<string, u_int32_t> > FloodMap;
	typedef FloodMap::iterator FloodIter;
	FloodMap seekers;
	FloodMap flooders;

	NmdcHub(const string& aHubURL);	
	virtual ~NmdcHub() throw();

	// Dummy
	NmdcHub(const NmdcHub&);
	NmdcHub& operator=(const NmdcHub&);

	void connect();

	void clearUsers();
	void ZLine(char* aLine, int iaLineLen) throw();
	void ChatLine(char* aLine) throw();
	void DcLine(char* aLine, int iaLineLen, char* bLine, int ibLineLen) throw();

	string fromNmdc(const string& str) const { return Text::acpToUtf8(str); }
	string toNmdc(const string& str) const { return Text::utf8ToAcp(str); }
	
	virtual string checkNick(const string& aNick);
	virtual string getHubURL();

	// TimerManagerListener
	virtual void on(TimerManagerListener::Second, u_int32_t aTick) throw();

	virtual void on(Connecting) throw() { fire(ClientListener::Connecting(), this); }
	virtual void on(Connected) throw() { lastActivity = GET_TICK(); fire(ClientListener::Connected(), this); }
	virtual void on(Line, const string& l) throw();
	virtual void on(Failed, const string&) throw();

};

#endif // NMDC_HUB_H

/* Some parts Copyright (C) 2002-2006 Petr Kozelka, PPK at czdc dot org */
